/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting;

import org.jboss.logging.Logger;
import org.jboss.remoting.loading.ClassByteClassLoader;
import org.jboss.remoting.marshal.MarshallLoaderFactory;
import org.jboss.remoting.security.SSLSocketBuilder;
import org.jboss.remoting.serialization.SerializationStreamFactory;

import javax.net.SocketFactory;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.HashMap;
import java.util.Map;

/**
 * AbstractInvoker is an abstract handler part that contains common methods between both
 * client and server.
 *
 * @author <a href="mailto:jhaynie@vocalocity.net">Jeff Haynie</a>
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 * @version $Revision: 1.12 $
 */
public abstract class AbstractInvoker implements Invoker
{   
   protected final Logger log = Logger.getLogger(getClass());
   protected ClassByteClassLoader classbyteloader;
   protected InvokerLocator locator;
   protected Map localServerLocators = new HashMap();
   protected String serializationType;
   protected Map configuration = new HashMap();
   protected SocketFactory socketFactory;
   
   // Indicates if the serverSocketFactory was generated internally.
   protected boolean socketFactoryCreatedFromSSLParameters;
   
   public AbstractInvoker(InvokerLocator locator)
   {
      this(locator, null);
   }
   
   public AbstractInvoker(InvokerLocator locator, Map configuration)
   {
      this.classbyteloader = new ClassByteClassLoader(this.getClass().getClassLoader());
      this.locator = locator;
      
      if (configuration != null)
         this.configuration.putAll(configuration);
      
      if (locator.getParameters() != null)
         this.configuration.putAll(locator.getParameters());

      try
      {
         InvokerLocator loaderLocator = MarshallLoaderFactory.convertLocator(locator);
         if(loaderLocator != null)
         {
            classbyteloader.setClientInvoker(new Client(loaderLocator));
         }
      }
      catch(Exception e)
      {
         log.error("Could not create remote class loading for invoker.", e);
      }

      if(locator == null || locator.getParameters() == null)
      {
         this.setSerializationType(SerializationStreamFactory.JAVA);
      }
      else
      {
         this.setSerializationType(locator.findSerializationType());
      }
   }

   /**
    * return the locator this Invoker represents
    *
    * @return
    */
   public InvokerLocator getLocator()
   {
      return locator;
   }

   /**
    * Sets the callback server locator for the specified callback listener id
    *
    * @param locator
    */
   public void addClientLocator(String listenerId, InvokerLocator locator)
   {
      Object obj = localServerLocators.put(listenerId, locator);
      if(obj != null)
      {
         throw new RuntimeException("InvokerLocator already exists for listener id " + listenerId + ".  " +
                                    "Remove this listener before adding again.");
      }
   }

   /**
    * Gets the client locator.  This locator will be used by the server side
    * to make callbacks to the handler for this locator.
    *
    * @return
    */
   public InvokerLocator getClientLocator(String listenerId)
   {
      return (InvokerLocator) localServerLocators.get(listenerId);
   }

   /**
    * set the classloader to use internally
    *
    * @param classloader
    */
   public synchronized void setClassLoader(ClassLoader classloader)
   {
      this.classbyteloader = new ClassByteClassLoader(classloader);
   }

   public synchronized ClassLoader getClassLoader()
   {
      return classbyteloader;
   }

   public String getSerializationType()
   {
      return serializationType;
   }

   public void setSerializationType(String serializationType)
   {
      this.serializationType = serializationType;
   }
   
   public SocketFactory getSocketFactory()
   {
      return socketFactory;
   }
   
   public void setSocketFactory(SocketFactory socketFactory)
   {
      this.socketFactory = socketFactory;
   }
   
   public boolean isSocketFactoryCreatedFromSSLParameters()
   {
      return socketFactoryCreatedFromSSLParameters;
   }
   
   /**
    * If any configuration parameters relate to the construction of a SSLSocketBuilder,
    * create one.
    *
    * @param configuration
    * @return
    */
   protected SocketFactory createSocketFactory(Map configuration)
   {
      if (configuration == null)
         return null;
      
      if (socketFactory != null)
         return socketFactory;
      
      SocketFactory factory = null;
      
      Object obj = configuration.get(Remoting.CUSTOM_SOCKET_FACTORY);
      if (obj != null)
      {
         if (obj instanceof SocketFactory)
         {
            factory = (SocketFactory) obj;
         }
         else
         {
            throw new RuntimeException("Can not set custom socket factory (" + obj + ") as is not of type javax.net.SocketFactory");
         }
      }
      
      if(factory == null)
      {
         String socketFactoryString = (String)configuration.get(Remoting.SOCKET_FACTORY_NAME);
         if(socketFactoryString != null && socketFactoryString.length() > 0)
         {
            //ClassLoader classLoader = invoker.getClassLoader();
            ClassLoader classLoader = null;
            if(classLoader == null)
            {
               classLoader = Thread.currentThread().getContextClassLoader();
               
               if(classLoader == null)
               {
                  classLoader = getClass().getClassLoader();
               }
            }
            try
            {
               Class cl = classLoader.loadClass(socketFactoryString);
               
               Constructor socketFactoryConstructor = null;
               socketFactoryConstructor = cl.getConstructor(new Class[]{});
               factory = (SocketFactory)socketFactoryConstructor.newInstance(new Object[] {});
               log.trace("SocketFactory (" + socketFactoryString + ") loaded");
            }
            catch(Exception e)
            {
               log.debug("Could not create socket factory by classname (" + socketFactoryString + ").  Error message: " + e.getMessage());
            }
         }
      }
      
      if (factory == null && needsCustomSSLConfiguration(configuration))
      {
         try
         {
            SSLSocketBuilder socketBuilder = new SSLSocketBuilder(configuration);
            socketBuilder.setUseSSLSocketFactory( false );
            factory = socketBuilder.createSSLSocketFactory();
            socketFactoryCreatedFromSSLParameters = true;
         }
         catch (IOException e)
         {
            throw new RuntimeException("Unable to create customized SSL socket factory", e);
         }
      }
      
      return factory;
   }

   
   public static boolean needsCustomSSLConfiguration(Map configuration)
   {
      if (configuration.get(SSLSocketBuilder.REMOTING_KEY_ALIAS) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_CLIENT_AUTH_MODE) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_SERVER_AUTH_MODE) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_SSL_PROTOCOL) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_SSL_PROVIDER_NAME) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_SERVER_SOCKET_USE_CLIENT_MODE) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_SOCKET_USE_CLIENT_MODE) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_KEY_PASSWORD) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_KEY_STORE_ALGORITHM) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_KEY_STORE_FILE_PATH) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_KEY_STORE_PASSWORD) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_KEY_STORE_TYPE) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_TRUST_STORE_ALGORITHM) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_TRUST_STORE_FILE_PATH) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_TRUST_STORE_PASSWORD) != null ||
            configuration.get(SSLSocketBuilder.REMOTING_TRUST_STORE_TYPE) != null
      )
         return true;
      else
         return false;
   }
}
