/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.remoting;

import java.util.Map;
import java.util.TimerTask;

import org.jboss.logging.Logger;
import org.jboss.remoting.util.TimerUtil;

/**
 * This class is used on the remoting server to maintain lease information
 * for remoting clients.  Will generate callback to ConnectionListener interface
 * if determined that client no longer available.
 *
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class Lease
{

   private ConnectionNotifier notifier = null;
   private String clientSessionId = null;
   private long leasePeriod = -1;
   private String locatorURL = null;
   private Map requestPayload = null;
   private LeaseTimerTask leaseTimerTask = null;
   private long leaseWindow = -1;
   private long pingStart = -1;
   private Map clientLeases = null;

   private boolean leaseUpdated = false;
   
   private static final Logger log = Logger.getLogger(Lease.class);


   public Lease(String clientSessionId, long leasePeriod, String locatorurl, Map requestPayload,
                ConnectionNotifier notifier, Map clientLeases)
   {
      this.clientSessionId = clientSessionId;
      this.leasePeriod = leasePeriod;
      this.notifier = notifier;
      this.locatorURL = locatorurl;
      this.requestPayload = requestPayload;
      this.leaseWindow = leasePeriod * 2;
      this.clientLeases = clientLeases;
   }


   public void startLease()
   {
      leaseTimerTask = new LeaseTimerTask();
      TimerUtil.schedule(leaseTimerTask, leaseWindow);
   }

   public void updateLease(long leasePeriod)
   {      
      leaseUpdated = true;      
      if (leasePeriod != this.leasePeriod)
      {
         this.leasePeriod = leasePeriod;
         this.leaseWindow = leasePeriod * 2;
         stopLease();
         startLease();
      }
      else
      {
         if (pingStart != -1)
         {
            long pingDuration = System.currentTimeMillis() - pingStart;
            if (pingDuration > 0.75 * leaseWindow)
            {
               leaseWindow = pingDuration * 2;

               stopLease();
               leaseTimerTask = new LeaseTimerTask();
               TimerUtil.schedule(leaseTimerTask, leaseWindow);           
            }
         }
         
      }
      pingStart = System.currentTimeMillis();
   }

   public void terminateLease(long leasePeriod)
   {
      stopLease();
      notifier.connectionTerminated(locatorURL, clientSessionId, requestPayload);
   }

   private void stopLease()
   {     
      leaseTimerTask.cancel();
   }

   private class LeaseTimerTask extends TimerTask
   {

      /**
       * The action to be performed by this timer task.
       */
      public void run()
      {
         if (leaseUpdated)
         {
            leaseUpdated = false;          
         }
         else
         {
            stopLease();
            notifier.connectionLost(locatorURL, clientSessionId, requestPayload);
            if (clientLeases != null)
            {
               clientLeases.remove(clientSessionId);
            }
         }
      }
   }
}