/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/
package org.jboss.remoting.callback;

import org.jboss.logging.Logger;
import org.jboss.remoting.Client;
import org.jboss.remoting.util.TimerUtil;

import java.util.List;
import java.util.Map;
import java.util.TimerTask;

/**
 * @author <a href="mailto:tom.elrod@jboss.com">Tom Elrod</a>
 */
public class CallbackPoller extends TimerTask
{
   /**
    * Default polling period for getting callbacks from the server.
    * Default is 5000 milliseconds.
    */
   public static final long DEFAULT_POLL_PERIOD = 5000;

   /**
    * The key value to use to specify the desired poll period
    * within the metadata Map.
    */
   public static final String CALLBACK_POLL_PERIOD = "callbackPollPeriod";

   private Client client = null;
   private InvokerCallbackHandler callbackHandler = null;
   private Map metadata = null;
   private Object callbackHandlerObject = null;
   private long pollPeriod = DEFAULT_POLL_PERIOD;

   private static final Logger log = Logger.getLogger(CallbackPoller.class);


   public CallbackPoller(Client client, InvokerCallbackHandler callbackhandler, Map metadata, Object callbackHandlerObject)
   {
      this.client = client;
      this.callbackHandler = callbackhandler;
      this.metadata = metadata;
      this.callbackHandlerObject = callbackHandlerObject;
   }

   public void start() throws Exception
   {
      if (callbackHandler == null)
      {
         throw new NullPointerException("Can not poll for callbacks when InvokerCallbackHandler is null.");
      }
      if (client != null)
      {
         client.connect();
      }

      if (metadata != null)
      {
         Object val = metadata.get(CALLBACK_POLL_PERIOD);
         if (val != null)
         {
            if (val instanceof String)
            {
               try
               {
                  pollPeriod = Long.parseLong((String) val);
               }
               catch (NumberFormatException e)
               {
                  log.warn("Error converting " + CALLBACK_POLL_PERIOD + " to type long.  " + e.getMessage());
               }
            }
            else
            {
               log.warn("Value for " + CALLBACK_POLL_PERIOD + " configuration must be of type " + String.class.getName() +
                        " and is " + val.getClass().getName());
            }
         }
      }

      TimerUtil.schedule(this, pollPeriod);

   }

   public void run()
   {
      // need to pull callbacks from server and give them to callback handler
      try
      {
         List callbacks = client.getCallbacks(callbackHandler);
         if (callbacks != null && callbacks.size() > 0)
         {
            final Callback[] callbackArray = (Callback[]) callbacks.toArray(new Callback[callbacks.size()]);

            // delivering callbacks on new thread as don't want to allow the callback handler
            // to hijack timer task thread as could slow down timer otherwise.
            new Thread(new Runnable()
            {
               public void run()
               {
                  for (int x = 0; x < callbackArray.length; x++)
                  {
                     Callback callbackObject = callbackArray[x];
                     callbackObject.setCallbackHandleObject(callbackHandlerObject);
                     try
                     {
                        callbackHandler.handleCallback(callbackObject);
                     }
                     catch (HandleCallbackException e)
                     {
                        log.error("Error delivering callback to callback handler (" + callbackHandler + ").", e);
                     }

                  }
               }
            }).start();
         }

      }
      catch (Throwable throwable)
      {
         log.error("Error getting callbacks from server.", throwable);
      }
   }

   public void stop()
   {
      cancel();
      client = null;
      callbackHandler = null;
      callbackHandlerObject = null;
   }
}