/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.rmi;

import org.jboss.remoting.Version;

import javax.net.SocketFactory;
import java.io.IOException;
import java.io.Serializable;
import java.net.Socket;
import java.rmi.server.RMIClientSocketFactory;


/**
 * <code>RemotingRMIClientSocketFactory</code> provides two services to <code>RMIServerInvoker</code>.
 * <ol>
 * <li>It can be parameterized by a host name, allowing <code>RMIServerInvoker</code> to supply RMI
 * with a factory which creates sockets connected to a specified host name.
 * <p/>
 * <li>It can be parameterized by a <code>SocketFactory</code> allowing <code>RMIServerInvoker</code>
 * to supply RMI with a factory facility which creates specialized sockets.
 * </ol>
 * <p/>
 * if the <code>SocketFactory</code> parameter is specified, then the <code>RemotingRMIClientSocketFactory</code>
 * should be used with a matching instance of <code>RemotingRMIServerSocketFactory</code> with a compatible
 * <code>ServerSocketFactory</code>.
 * <p/>
 * If the <code>SocketFactory</code> parameter is not specified, an instance of <code>java.net.Socket</code>
 * will be created by default.
 * <p/>
 * Although there is no apparent need for the host name parameter, since the <code>createSocket()</code>
 * method receives a host name, it seems that for a server object bound to localhost, the RMI runtime will
 * pass to <code>createSocket()</code> one of the IP addresses for which the host is configured (other than
 * 127.0.0.1), resulting in a failure to retrieve the object from the Registry.  If a host name is passed to
 * a <code>RemotingRMIClientFactory</code> constructor, it will override the host name passed to
 * <code>createSocket()</code>  In particular, parameterizing <code>RemotingRMIClientSocketFactory</code>
 * with localhost will allow the retrieval of objects bound to localhost.
 *
 * @author <a href="mailto:r.sigal@computer.org">Ron Sigal</a>
 * @version $Revision: 1.9 $
 *          <p/>
 *          Copyright (c) 2005
 *          </p>
 */

public class RemotingRMIClientSocketFactory implements RMIClientSocketFactory, Serializable
{
   static final long serialVersionUID;

   transient protected SocketFactory socketFactory;
   private int timeout = 60000;

   // The commented code below is from an attempt to incorporate a <code>java.lang.reflect.Constructor</code>
   // parameter to provide a very general way to create sockets.  The problem is that
   // <code>Constructor</code> does not implement <code>Serializable</code>, which is necessary to
   // allow the <code>RemotingRMIClientSocketFactory</code> to be transmitted to the client.  The
   // code is left in place because it could be resurrected by passing in a class name and parameter
   // types to specify a constructor.  Fortunately, <code>java.lang.Class</code> does implement
   // <code>Serializable</code>.

//   private Constructor constructor;
//   private Object[] args;
//   private int hostPosition;
//   private int portPosition;

   protected String hostName;

   private static int defaultHashCode = "RemotingRMISocketFactory".hashCode();

   static
   {
      if(Version.getDefaultVersion() == Version.VERSION_1)
      {
         serialVersionUID = -7491556589517716155L;
      }
      else
      {
         serialVersionUID = -3039839695840773968L;
      }
   }

   /**
    */
   public RemotingRMIClientSocketFactory()
   {
   }


   /**
    * @param socketFactory the <code>SocketFactory</code> to be used to generate sockets
    */
   public RemotingRMIClientSocketFactory(SocketFactory socketFactory)
   {
      this.socketFactory = socketFactory;
   }


   /**
    * @param hostName name of host to which all generated sockets should be connected
    */
   public RemotingRMIClientSocketFactory(String hostName, int timeout)
   {
      this.hostName = hostName;
      this.timeout = timeout;
   }

   public RemotingRMIClientSocketFactory(SocketFactory socketFactory, String hostName, int timeout)
   {
      this.socketFactory = socketFactory;
      this.hostName = hostName;
      this.timeout = timeout;
   }


   /**
    * @param socketFactory the <code>SocketFactory</code> to be used to generate sockets
    * @param hostName      name of host to which all generated sockets should be connected
    */
   public RemotingRMIClientSocketFactory(SocketFactory socketFactory, String hostName)
   {
      this.hostName = hostName;
      this.socketFactory = socketFactory;
   }

//   public RemotingRMISocketFactory(Constructor constructor, Object[] args, int hostPosition, int portPosition)
//   throws ClassNotFoundException, NoSuchMethodException
//   {
//      this.constructor = constructor;
//      this.args = args;
//      this.portPosition = portPosition;
//   }


   /**
    * Creates a new socket.  If a <code>SocketFactory</code> was passed to the constructor, it will be used.
    * Otherwise, a <code>java.net.Socket</code> will be created by default.
    *
    * @param host host to which socket should be connected
    * @param port port to which socket should be connected
    * @return new socket
    * @throws IOException if there is a problem creating a socket
    */
   public Socket createSocket(String host, int port) throws IOException
   {
      String effectiveHost = hostName != null ? hostName : host;

      Socket socket = null;
      if(socketFactory != null)
      {
         socket = socketFactory.createSocket(effectiveHost, port);
      }

//      if (constructor != null)
//      {
//         if (hostPosition != -1)
//            args[hostPosition] = host;
//
//         if (portPosition != -1)
//            args[portPosition] = new Integer(port);
//
//         try
//         {
//            return (Socket) constructor.newInstance(args);
//         }
//         catch (Exception e)
//         {
//            throw new IOException(e.getMessage());
//         }
//      }

      else
      {
         socket = new Socket(effectiveHost, port);
      }

      socket.setSoTimeout(timeout);
      return socket;
   }


   /**
    * Overrides <code>equals()</code> method provided by <code>Object</code> class.  It looks for
    * equality of host name and socket factory parameters.
    *
    * @param o <code>Object</code> to which <code>this</code> is to be compared
    * @return true if and only if <code>o</code> and <code>this</code> are equal
    */
   public boolean equals(Object o)
   {
      if(! (o instanceof RemotingRMIClientSocketFactory))
      {
         return false;
      }

      RemotingRMIClientSocketFactory sf = (RemotingRMIClientSocketFactory) o;

      // This is for the version that uses a SocketFactory
      if(socketFactory != null)
      {
         if(sf.socketFactory == null || ! socketFactory.equals(sf.socketFactory))
         {
            return false;
         }

         if(hostName != null)
         {
            if(sf.hostName == null || ! hostName.equals(sf.hostName))
            {
               return false;
            }
            else
            {
               return true;
            }
         }

         if(sf.hostName != null)
         {
            return false;
         }
         else
         {
            return true;
         }
      }

//      // This is for the version that uses a constructor
//      if (constructor != null)
//      {
//         if (sf.constructor == null ||
//             ! constructor.equals(sf.constructor) ||
//             hostPosition != sf.hostPosition ||
//             portPosition != sf.portPosition)
//            return false;
//
//         for (int i = 0; i < args.length; i++)
//         {
//            if (!args[i].equals(sf.args[i]))
//               return false;
//         }
//
//         return true;
//      }

      // This is for the plain vanilla version.
      if(sf.socketFactory != null)
      {
         return false;
      }

      if(hostName != null)
      {
         if(sf.hostName == null || ! hostName.equals(sf.hostName))
         {
            return false;
         }
         else
         {
            return true;
         }
      }

      if(sf.hostName != null)
      {
         return false;
      }
      else
      {
         return true;
      }
   }


   /**
    * Overrides <code>hashCode()</code> method provided by the <code>Object</code> class.
    *
    * @return a hashcode for <code>this</code>
    */
   public int hashCode()
   {
      if(socketFactory != null)
      {
         if(hostName != null)
         {
            return socketFactory.hashCode() * hostName.hashCode();
         }
         else
         {
            return socketFactory.hashCode();
         }
      }

//      if (constructor != null)
//      {
//         int hash = hostPosition * portPosition;
//         
//         for (int i = 0; i < args.length; i++)
//            hash *= args[i].hashCode();
//         
//         return hash;
//      }

      if(hostName != null)
      {
         return hostName.hashCode();
      }

      return defaultHashCode;
   }
}